@@ -18,7 +18,6 @@ import difflib
 import functools
 import importlib.machinery
 import io
-import itertools
 import json
 import os
 import pathlib
@@ -31,6 +30,7 @@ import sysconfig
 import tarfile
 import tempfile
 import textwrap
+import time
 import typing
 import warnings
 
@@ -46,8 +46,7 @@ import packaging.version
 import pyproject_metadata
 
 import mesonpy._compat
-import mesonpy._dylib
-import mesonpy._elf
+import mesonpy._rpath
 import mesonpy._tags
 import mesonpy._util
 import mesonpy._wheelfile
@@ -56,9 +55,7 @@ from mesonpy._compat import Collection, Mapping, cached_property, read_binary
 
 
 if typing.TYPE_CHECKING:  # pragma: no cover
-    from typing import (
-        Any, Callable, ClassVar, DefaultDict, List, Literal, Optional, Sequence, TextIO, Tuple, Type, TypeVar, Union
-    )
+    from typing import Any, Callable, DefaultDict, List, Literal, Optional, Sequence, TextIO, Tuple, Type, TypeVar, Union
 
     from mesonpy._compat import Iterator, ParamSpec, Path
 
@@ -66,7 +63,7 @@ if typing.TYPE_CHECKING:  # pragma: no cover
     T = TypeVar('T')
 
 
-__version__ = '0.13.0'
+__version__ = '0.14.0'
 
 
 # XXX: Once Python 3.8 is our minimum supported version, get rid of
@@ -128,9 +125,8 @@ _STYLES = _init_colors()  # holds the color values, should be _COLORS or _NO_COL
 
 
 _SUFFIXES = importlib.machinery.all_suffixes()
-_EXTENSION_SUFFIXES = importlib.machinery.EXTENSION_SUFFIXES
-_EXTENSION_SUFFIX_REGEX = re.compile(r'^\.(?:(?P<abi>[^.]+)\.)?(?:so|pyd|dll)$')
-assert all(re.match(_EXTENSION_SUFFIX_REGEX, x) for x in _EXTENSION_SUFFIXES)
+_EXTENSION_SUFFIX_REGEX = re.compile(r'^[^.]+\.(?:(?P<abi>[^.]+)\.)?(?:so|pyd|dll)$')
+assert all(re.match(_EXTENSION_SUFFIX_REGEX, f'foo{x}') for x in importlib.machinery.EXTENSION_SUFFIXES)
 
 
 # Map Meson installation path placeholders to wheel installation paths.
@@ -153,10 +149,11 @@ def _map_to_wheel(sources: Dict[str, Dict[str, Any]]) -> DefaultDict[str, List[T
     wheel_files: DefaultDict[str, List[Tuple[pathlib.Path, str]]] = collections.defaultdict(list)
     packages: Dict[str, str] = {}
 
-    for group in sources.values():
+    for key, group in sources.items():
         for src, target in group.items():
             destination = pathlib.Path(target['destination'])
             anchor = destination.parts[0]
+            dst = pathlib.Path(*destination.parts[1:])
 
             path = _INSTALLATION_PATH_MAP.get(anchor)
             if path is None:
@@ -170,9 +167,31 @@ def _map_to_wheel(sources: Dict[str, Dict[str, Any]]) -> DefaultDict[str, List[T
                     that = os.fspath(other / next(d for d, s in wheel_files[other] if d.parts[0] == destination.parts[1]))
                     raise BuildError(
                         f'The {package} package is split between {path} and {other}: '
-                        f'{this!r} and {that!r}, a "pure: false" argument may be missing in meson.build')
+                        f'{this!r} and {that!r}, a "pure: false" argument may be missing in meson.build. '
+                        f'It is recommended to set it in "import(\'python\').find_installation()"')
+
+            if key == 'install_subdirs':
+                assert os.path.isdir(src)
+                exclude_files = {os.path.normpath(x) for x in target.get('exclude_files', [])}
+                exclude_dirs = {os.path.normpath(x) for x in target.get('exclude_dirs', [])}
+                for root, dirnames, filenames in os.walk(src):
+                    for name in dirnames.copy():
+                        dirsrc = os.path.join(root, name)
+                        relpath = os.path.relpath(dirsrc, src)
+                        if relpath in exclude_dirs:
+                            dirnames.remove(name)
+                    # sort to process directories determninistically
+                    dirnames.sort()
+                    for name in sorted(filenames):
+                        filesrc = os.path.join(root, name)
+                        relpath = os.path.relpath(filesrc, src)
+                        if relpath in exclude_files:
+                            continue
+                        filedst = dst / relpath
+                        wheel_files[path].append((filedst, filesrc))
+            else:
+                wheel_files[path].append((dst, src))
 
-            wheel_files[path].append((pathlib.Path(*destination.parts[1:]), src))
     return wheel_files
 
 
@@ -219,13 +238,42 @@ class MesonBuilderError(Error):
     """Error when building the Meson package."""
 
 
+class Metadata(pyproject_metadata.StandardMetadata):
+    # The class method from the pyproject_metadata base class is not
+    # typed in a subclassing friendly way, thus annotations to ignore
+    # typing are needed.
+
+    @classmethod
+    def from_pyproject(cls, data: Mapping[str, Any], project_dir: Path) -> Metadata:  # type: ignore[override]
+        metadata = super().from_pyproject(data, project_dir)
+
+        # Check for missing version field.
+        if not metadata.version and 'version' not in metadata.dynamic:
+            raise pyproject_metadata.ConfigurationError(
+                'Required "project.version" field is missing and not declared as dynamic')
+
+        # Check for unsupported dynamic fields.
+        unsupported_dynamic = set(metadata.dynamic) - {'version', }
+        if unsupported_dynamic:
+            fields = ', '.join(f'"{x}"' for x in unsupported_dynamic)
+            raise pyproject_metadata.ConfigurationError(f'Unsupported dynamic fields: {fields}')
+
+        return metadata  # type: ignore[return-value]
+
+    # Local fix for a bug in pyproject-metadata. See
+    # https://github.com/mesonbuild/meson-python/issues/454
+    def _update_dynamic(self, value: Any) -> None:
+        if value and 'version' in self.dynamic:
+            self.dynamic.remove('version')
+
+
 class _WheelBuilder():
     """Helper class to build wheels from projects."""
 
     def __init__(
         self,
         project: Project,
-        metadata: Optional[pyproject_metadata.StandardMetadata],
+        metadata: Metadata,
         source_dir: pathlib.Path,
         build_dir: pathlib.Path,
         sources: Dict[str, Dict[str, Any]],
@@ -236,20 +284,18 @@ class _WheelBuilder():
         self._build_dir = build_dir
         self._sources = sources
 
-        self._libs_build_dir = self._build_dir / 'mesonpy-wheel-libs'
-
     @cached_property
     def _wheel_files(self) -> DefaultDict[str, List[Tuple[pathlib.Path, str]]]:
         return _map_to_wheel(self._sources)
 
     @property
     def _has_internal_libs(self) -> bool:
-        return bool(self._wheel_files['mesonpy-libs'])
+        return bool(self._wheel_files.get('mesonpy-libs'))
 
     @property
     def _has_extension_modules(self) -> bool:
         # Assume that all code installed in {platlib} is Python ABI dependent.
-        return bool(self._wheel_files['platlib'])
+        return bool(self._wheel_files.get('platlib'))
 
     @property
     def normalized_name(self) -> str:
@@ -257,11 +303,8 @@ class _WheelBuilder():
 
     @property
     def basename(self) -> str:
-        """Normalized wheel name and version (eg. meson_python-1.0.0)."""
-        return '{distribution}-{version}'.format(
-            distribution=self.normalized_name,
-            version=self._project.version,
-        )
+        """Normalized wheel name and version."""
+        return f'{self.normalized_name}-{self._project.version}'
 
     @property
     def tag(self) -> mesonpy._tags.Tag:
@@ -279,10 +322,7 @@ class _WheelBuilder():
     @property
     def name(self) -> str:
         """Wheel name, this includes the basename and tag."""
-        return '{basename}-{tag}'.format(
-            basename=self.basename,
-            tag=self.tag,
-        )
+        return f'{self.basename}-{self.tag}'
 
     @property
     def distinfo_dir(self) -> str:
@@ -322,9 +362,6 @@ class _WheelBuilder():
     @property
     def entrypoints_txt(self) -> bytes:
         """dist-info entry_points.txt."""
-        if not self._metadata:
-            return b''
-
         data = self._metadata.entrypoints.copy()
         data.update({
             'console_scripts': self._metadata.scripts,
@@ -343,34 +380,20 @@ class _WheelBuilder():
 
     @cached_property
     def _stable_abi(self) -> Optional[str]:
-        """Determine stabe ABI compatibility.
-
-        Examine all files installed in {platlib} that look like
-        extension modules (extension .pyd on Windows, .dll on Cygwin,
-        and .so on other platforms) and, if they all share the same
-        PEP 3149 filename stable ABI tag, return it.
-
-        Other files are ignored.
-
-        """
-        soext = sorted(_EXTENSION_SUFFIXES, key=len)[0]
-        abis = []
-
-        for path, _ in self._wheel_files['platlib']:
-            # NOTE: When searching for shared objects files, we assume the host
-            # and build machines have the same soext, even though that we might
-            # be cross compiling.
-            if path.suffix == soext:
-                match = re.match(r'^[^.]+(.*)$', path.name)
-                assert match is not None
-                suffix = match.group(1)
-                match = _EXTENSION_SUFFIX_REGEX.match(suffix)
+        if self._project._limited_api:
+            # Verify stabe ABI compatibility: examine files installed
+            # in {platlib} that look like extension modules, and raise
+            # an exception if any of them has a Python version
+            # specific extension filename suffix ABI tag.
+            for path, _ in self._wheel_files['platlib']:
+                match = _EXTENSION_SUFFIX_REGEX.match(path.name)
                 if match:
-                    abis.append(match.group('abi'))
-
-        stable = [x for x in abis if x and re.match(r'abi\d+', x)]
-        if len(stable) > 0 and len(stable) == len(abis) and all(x == stable[0] for x in stable[1:]):
-            return stable[0]
+                    abi = match.group('abi')
+                    if abi is not None and abi != 'abi3':
+                        raise BuildError(
+                            f'The package declares compatibility with Python limited API but extension '
+                            f'module {os.fspath(path)!r} is tagged for a specific Python version.')
+            return 'abi3'
         return None
 
     @property
@@ -394,16 +417,16 @@ class _WheelBuilder():
         self._project.build()  # the project needs to be built for this :/
 
         with open(file, 'rb') as f:
-            if platform.system() == 'Linux':
+            if sys.platform == 'linux':
                 return f.read(4) == b'\x7fELF'  # ELF
-            elif platform.system() == 'Darwin':
+            elif sys.platform == 'darwin':
                 return f.read(4) in (
                     b'\xfe\xed\xfa\xce',  # 32-bit
                     b'\xfe\xed\xfa\xcf',  # 64-bit
                     b'\xcf\xfa\xed\xfe',  # arm64
                     b'\xca\xfe\xba\xbe',  # universal / fat (same as java class so beware!)
                 )
-            elif platform.system() == 'Windows':
+            elif sys.platform == 'win32':
                 return f.read(2) == b'MZ'
 
         # For unknown platforms, check for file extensions.
@@ -412,67 +435,26 @@ class _WheelBuilder():
             return True
         return False
 
-    def _install_path(  # noqa: C901
-        self,
-        wheel_file: mesonpy._wheelfile.WheelFile,
-        counter: mesonpy._util.CLICounter,
-        origin: Path,
-        destination: pathlib.Path,
-    ) -> None:
-        """"Install" file or directory into the wheel
-        and do the necessary processing before doing so.
+    def _install_path(self, wheel_file: mesonpy._wheelfile.WheelFile, origin: Path, destination: pathlib.Path) -> None:
+        """Add a file to the wheel."""
 
-        Some files might need to be fixed up to set the RPATH to the internal
-        library directory on Linux wheels for eg.
-        """
-        location = destination.as_posix()
-        counter.update(location)
+        if self._has_internal_libs:
+            if self._is_native(os.fspath(origin)):
+                # When an executable, libray, or Python extension module is
+                # dynamically linked to a library built as part of the project,
+                # Meson adds a library load path to it pointing to the build
+                # directory, in the form of a relative RPATH entry. meson-python
+                # relocates the shared libraries to the $project.mesonpy.libs
+                # folder. Rewrite the RPATH to point to that folder instead.
+                libspath = os.path.relpath(f'.{self._project.name}.mesonpy.libs', destination.parent)
+                mesonpy._rpath.fix_rpath(origin, libspath)
 
-        # fix file
-        if os.path.isdir(origin):
-            for root, dirnames, filenames in os.walk(str(origin)):
-                # Sort the directory names so that `os.walk` will walk them in a
-                # defined order on the next iteration.
-                dirnames.sort()
-                for name in sorted(filenames):
-                    path = os.path.normpath(os.path.join(root, name))
-                    if os.path.isfile(path):
-                        arcname = os.path.join(destination, os.path.relpath(path, origin).replace(os.path.sep, '/'))
-                        wheel_file.write(path, arcname)
-        else:
-            if self._has_internal_libs:
-                if platform.system() == 'Linux' or platform.system() == 'Darwin':
-                    # add .mesonpy.libs to the RPATH of ELF files
-                    if self._is_native(os.fspath(origin)):
-                        # copy ELF to our working directory to avoid Meson having to regenerate the file
-                        new_origin = self._libs_build_dir / pathlib.Path(origin).relative_to(self._build_dir)
-                        os.makedirs(new_origin.parent, exist_ok=True)
-                        shutil.copy2(origin, new_origin)
-                        origin = new_origin
-                        # add our in-wheel libs folder to the RPATH
-                        if platform.system() == 'Linux':
-                            elf = mesonpy._elf.ELF(origin)
-                            libdir_path = \
-                                f'$ORIGIN/{os.path.relpath(f".{self._project.name}.mesonpy.libs", destination.parent)}'
-                            if libdir_path not in elf.rpath:
-                                elf.rpath = [*elf.rpath, libdir_path]
-                        elif platform.system() == 'Darwin':
-                            dylib = mesonpy._dylib.Dylib(origin)
-                            libdir_path = \
-                                f'@loader_path/{os.path.relpath(f".{self._project.name}.mesonpy.libs", destination.parent)}'
-                            if libdir_path not in dylib.rpath:
-                                dylib.rpath = [*dylib.rpath, libdir_path]
-                        else:
-                            # Internal libraries are currently unsupported on this platform
-                            raise NotImplementedError("Bundling libraries in wheel is not supported on platform '{}'"
-                                                      .format(platform.system()))
-
-            try:
-                wheel_file.write(origin, location)
-            except FileNotFoundError:
-                # work around for Meson bug, see https://github.com/mesonbuild/meson/pull/11655
-                if not os.fspath(origin).endswith('.pdb'):
-                    raise
+        try:
+            wheel_file.write(origin, destination.as_posix())
+        except FileNotFoundError:
+            # work around for Meson bug, see https://github.com/mesonbuild/meson/pull/11655
+            if not os.fspath(origin).endswith('.pdb'):
+                raise
 
     def _wheel_write_metadata(self, whl: mesonpy._wheelfile.WheelFile) -> None:
         # add metadata
@@ -491,35 +473,28 @@ class _WheelBuilder():
     def build(self, directory: Path) -> pathlib.Path:
         # ensure project is built
         self._project.build()
-        # install the project
-        self._project.install()
 
         wheel_file = pathlib.Path(directory, f'{self.name}.whl')
-
         with mesonpy._wheelfile.WheelFile(wheel_file, 'w') as whl:
             self._wheel_write_metadata(whl)
 
-            print('{light_blue}{bold}Copying files to wheel...{reset}'.format(**_STYLES))
-            with mesonpy._util.cli_counter(
-                len(list(itertools.chain.from_iterable(self._wheel_files.values()))),
-            ) as counter:
-                # install root scheme files
-                root_scheme = 'purelib' if self.is_pure else 'platlib'
-                for destination, origin in self._wheel_files[root_scheme]:
-                    self._install_path(whl, counter, origin, destination)
+            with mesonpy._util.cli_counter(sum(len(x) for x in self._wheel_files.values())) as counter:
 
-                # install bundled libraries
-                for destination, origin in self._wheel_files['mesonpy-libs']:
-                    destination = pathlib.Path(f'.{self._project.name}.mesonpy.libs', destination)
-                    self._install_path(whl, counter, origin, destination)
+                root = 'purelib' if self.is_pure else 'platlib'
 
-                # install the other schemes
-                for scheme in self._wheel_files.keys():
-                    if scheme in (root_scheme, 'mesonpy-libs'):
-                        continue
-                    for destination, origin in self._wheel_files[scheme]:
-                        destination = pathlib.Path(self.data_dir, scheme, destination)
-                        self._install_path(whl, counter, origin, destination)
+                for path, entries in self._wheel_files.items():
+                    for dst, src in entries:
+                        counter.update(src)
+
+                        if path == root:
+                            pass
+                        elif path == 'mesonpy-libs':
+                            # custom installation path for bundled libraries
+                            dst = pathlib.Path(f'.{self._project.name}.mesonpy.libs', dst)
+                        else:
+                            dst = pathlib.Path(self.data_dir, path, dst)
+
+                        self._install_path(whl, src, dst)
 
         return wheel_file
 
@@ -576,10 +551,16 @@ def _validate_pyproject_config(pyproject: Dict[str, Any]) -> Dict[str, Any]:
             raise ConfigError(f'Configuration entry "{name}" must be a list of strings')
         return value
 
+    def _bool(value: Any, name: str) -> bool:
+        if not isinstance(value, bool):
+            raise ConfigError(f'Configuration entry "{name}" must be a boolean')
+        return value
+
     scheme = _table({
+        'limited-api': _bool,
         'args': _table({
             name: _strings for name in _MESON_ARGS_KEYS
-        })
+        }),
     })
 
     table = pyproject.get('tool', {}).get('meson-python', {})
@@ -627,46 +608,41 @@ def _validate_config_settings(config_settings: Dict[str, Any]) -> Dict[str, Any]
 class Project():
     """Meson project wrapper to generate Python artifacts."""
 
-    _ALLOWED_DYNAMIC_FIELDS: ClassVar[List[str]] = [
-        'version',
-    ]
-    _metadata: pyproject_metadata.StandardMetadata
-
-    def __init__(
+    def __init__(  # noqa: C901
         self,
         source_dir: Path,
-        working_dir: Path,
-        build_dir: Optional[Path] = None,
+        build_dir: Path,
         meson_args: Optional[MesonArgs] = None,
         editable_verbose: bool = False,
     ) -> None:
         self._source_dir = pathlib.Path(source_dir).absolute()
-        self._working_dir = pathlib.Path(working_dir).absolute()
-        self._build_dir = pathlib.Path(build_dir).absolute() if build_dir else (self._working_dir / 'build')
+        self._build_dir = pathlib.Path(build_dir).absolute()
         self._editable_verbose = editable_verbose
-        self._install_dir = self._working_dir / 'install'
         self._meson_native_file = self._build_dir / 'meson-python-native-file.ini'
         self._meson_cross_file = self._build_dir / 'meson-python-cross-file.ini'
         self._meson_args: MesonArgs = collections.defaultdict(list)
-        self._env = os.environ.copy()
+        self._limited_api = False
 
         _check_meson_version()
 
         self._ninja = _env_ninja_command()
         if self._ninja is None:
             raise ConfigError(f'Could not find ninja version {_NINJA_REQUIRED_VERSION} or newer.')
-        self._env.setdefault('NINJA', self._ninja)
+        os.environ.setdefault('NINJA', self._ninja)
+
+        # make sure the build dir exists
+        self._build_dir.mkdir(exist_ok=True, parents=True)
 
         # setuptools-like ARCHFLAGS environment variable support
         if sysconfig.get_platform().startswith('macosx-'):
-            archflags = self._env.get('ARCHFLAGS', '').strip()
+            archflags = os.environ.get('ARCHFLAGS', '').strip()
             if archflags:
                 arch, *other = filter(None, (x.strip() for x in archflags.split('-arch')))
                 if other:
                     raise ConfigError(f'Multi-architecture builds are not supported but $ARCHFLAGS={archflags!r}')
                 macver, _, nativearch = platform.mac_ver()
                 if arch != nativearch:
-                    x = self._env.setdefault('_PYTHON_HOST_PLATFORM', f'macosx-{macver}-{arch}')
+                    x = os.environ.setdefault('_PYTHON_HOST_PLATFORM', f'macosx-{macver}-{arch}')
                     if not x.endswith(arch):
                         raise ConfigError(f'$ARCHFLAGS={archflags!r} and $_PYTHON_HOST_PLATFORM={x!r} do not agree')
                     family = 'aarch64' if arch == 'arm64' else arch
@@ -674,8 +650,10 @@ class Project():
                         [binaries]
                         c = ['cc', '-arch', {arch!r}]
                         cpp = ['c++', '-arch', {arch!r}]
+                        objc = ['cc', '-arch', {arch!r}]
+                        objcpp = ['c++', '-arch', {arch!r}]
                         [host_machine]
-                        system = 'Darwin'
+                        system = 'darwin'
                         cpu = {arch!r}
                         cpu_family = {family!r}
                         endian = 'little'
@@ -697,10 +675,6 @@ class Project():
             for key, value in meson_args.items():
                 self._meson_args[key].extend(value)
 
-        # make sure the build dir exists
-        self._build_dir.mkdir(exist_ok=True, parents=True)
-        self._install_dir.mkdir(exist_ok=True, parents=True)
-
         # write the native file
         native_file_data = textwrap.dedent(f'''
             [binaries]
@@ -721,20 +695,46 @@ class Project():
 
         # package metadata
         if 'project' in pyproject:
-            self._metadata = pyproject_metadata.StandardMetadata.from_pyproject(pyproject, self._source_dir)
+            self._metadata = Metadata.from_pyproject(pyproject, self._source_dir)
+            # set version from meson.build if version is declared as dynamic
+            if 'version' in self._metadata.dynamic:
+                version = self._meson_version
+                if version == 'undefined':
+                    raise pyproject_metadata.ConfigurationError(
+                        'Field "version" declared as dynamic but version is not defined in meson.build')
+                self._metadata.version = packaging.version.Version(version)
         else:
-            self._metadata = pyproject_metadata.StandardMetadata(
-                name=self._meson_name, version=packaging.version.Version(self._meson_version))
-        self._validate_metadata()
+            # if project section is missing, use minimal metdata from meson.build
+            name, version = self._meson_name, self._meson_version
+            if version == 'undefined':
+                raise pyproject_metadata.ConfigurationError(
+                    'Section "project" missing in pyproject.toml and version is not defined in meson.build')
+            self._metadata = Metadata(name=name, version=packaging.version.Version(version))
 
-        # set version from meson.build if dynamic
-        if 'version' in self._metadata.dynamic:
-            self._metadata.version = packaging.version.Version(self._meson_version)
+        # verify that we are running on a supported interpreter
+        if self._metadata.requires_python:
+            self._metadata.requires_python.prereleases = True
+            if platform.python_version().rstrip('+') not in self._metadata.requires_python:
+                raise MesonBuilderError(
+                    f'Package requires Python version {self._metadata.requires_python}, '
+                    f'running on {platform.python_version()}')
+
+        # limited API
+        self._limited_api = pyproject_config.get('limited-api', False)
+        if self._limited_api:
+            # check whether limited API is disabled for the Meson project
+            options = self._info('intro-buildoptions')
+            value = next((option['value'] for option in options if option['name'] == 'python.allow_limited_api'), None)
+            if not value:
+                self._limited_api = False
 
     def _run(self, cmd: Sequence[str]) -> None:
         """Invoke a subprocess."""
-        print('{cyan}{bold}+ {}{reset}'.format(' '.join(cmd), **_STYLES))
-        r = subprocess.run(cmd, env=self._env, cwd=self._build_dir)
+        # Flush the line to ensure that the log line with the executed
+        # command line appears before the command output. Without it,
+        # the lines appear in the wrong order in pip output.
+        print('{cyan}{bold}+ {}{reset}'.format(' '.join(cmd), **_STYLES), flush=True)
+        r = subprocess.run(cmd, cwd=self._build_dir)
         if r.returncode != 0:
             raise SystemExit(r.returncode)
 
@@ -759,27 +759,6 @@ class Project():
 
         self._run(['meson', 'setup', *setup_args])
 
-    def _validate_metadata(self) -> None:
-        """Check the pyproject.toml metadata and see if there are any issues."""
-
-        # check for unsupported dynamic fields
-        unsupported_dynamic = {
-            key for key in self._metadata.dynamic
-            if key not in self._ALLOWED_DYNAMIC_FIELDS
-        }
-        if unsupported_dynamic:
-            s = ', '.join(f'"{x}"' for x in unsupported_dynamic)
-            raise MesonBuilderError(f'Unsupported dynamic fields: {s}')
-
-        # check if we are running on an unsupported interpreter
-        if self._metadata.requires_python:
-            self._metadata.requires_python.prereleases = True
-            if platform.python_version().rstrip('+') not in self._metadata.requires_python:
-                raise MesonBuilderError(
-                    f'Unsupported Python version {platform.python_version()}, '
-                    f'expected {self._metadata.requires_python}'
-                )
-
     @cached_property
     def _wheel_builder(self) -> _WheelBuilder:
         return _WheelBuilder(
@@ -793,7 +772,7 @@ class Project():
     @property
     def _build_command(self) -> List[str]:
         assert self._ninja is not None  # help mypy out
-        if platform.system() == 'Windows':
+        if sys.platform == 'win32':
             # On Windows use 'meson compile' to setup the MSVC compiler
             # environment. Using the --ninja-args option allows to
             # provide the exact same semantics for the compile arguments
@@ -810,55 +789,38 @@ class Project():
         """Build the Meson project."""
         self._run(self._build_command)
 
-    def install(self) -> None:
-        """Install the Meson project."""
-        destdir = os.fspath(self._install_dir)
-        self._run(['meson', 'install', '--no-rebuild', '--destdir', destdir, *self._meson_args['install']])
-
-    @classmethod
-    @contextlib.contextmanager
-    def with_temp_working_dir(
-        cls,
-        source_dir: Path = os.path.curdir,
-        build_dir: Optional[Path] = None,
-        meson_args: Optional[MesonArgs] = None,
-        editable_verbose: bool = False,
-    ) -> Iterator[Project]:
-        """Creates a project instance pointing to a temporary working directory."""
-        with tempfile.TemporaryDirectory(prefix='.mesonpy-', dir=os.fspath(source_dir)) as tmpdir:
-            yield cls(source_dir, tmpdir, build_dir, meson_args, editable_verbose)
-
     @functools.lru_cache()
-    def _info(self, name: str) -> Dict[str, Any]:
+    def _info(self, name: str) -> Any:
         """Read info from meson-info directory."""
-        file = self._build_dir.joinpath('meson-info', f'{name}.json')
-        return typing.cast(
-            Dict[str, str],
-            json.loads(file.read_text())
-        )
+        info = self._build_dir.joinpath('meson-info', f'{name}.json')
+        return json.loads(info.read_text())
 
     @property
     def _install_plan(self) -> Dict[str, Dict[str, Dict[str, str]]]:
         """Meson install_plan metadata."""
+        install_plan = self._info('intro-install_plan')
 
-        # copy the install plan so we can modify it
-        install_plan = self._info('intro-install_plan').copy()
-
-        # parse install args for install tags (--tags)
+        # parse install args to extract --tags and --skip-subprojects
         parser = argparse.ArgumentParser()
         parser.add_argument('--tags')
+        parser.add_argument('--skip-subprojects', nargs='?', const='*', default='')
         args, _ = parser.parse_known_args(self._meson_args['install'])
+        install_tags = {t.strip() for t in args.tags.split(',')} if args.tags else None
+        skip_subprojects = {p for p in (p.strip() for p in args.skip_subprojects.split(',')) if p}
 
-        # filter the install_plan for files that do not fit the install tags
-        if args.tags:
-            install_tags = args.tags.split(',')
+        manifest: DefaultDict[str, Dict[str, Dict[str, str]]] = collections.defaultdict(dict)
 
-            for files in install_plan.values():
-                for file, details in list(files.items()):
-                    if details['tag'].strip() not in install_tags:
-                        del files[file]
+        # filter install_plan accordingly
+        for key, targets in install_plan.items():
+            for target, details in targets.items():
+                if install_tags is not None and details['tag'] not in install_tags:
+                    continue
+                subproject = details.get('subproject')
+                if subproject is not None and (subproject in skip_subprojects or '*' in skip_subprojects):
+                    continue
+                manifest[key][target] = details
 
-        return install_plan
+        return manifest
 
     @property
     def _meson_name(self) -> str:
@@ -891,10 +853,9 @@ class Project():
 
     @property
     def license_file(self) -> Optional[pathlib.Path]:
-        if self._metadata:
-            license_ = self._metadata.license
-            if license_ and license_.file:
-                return pathlib.Path(license_.file)
+        license_ = self._metadata.license
+        if license_ and license_.file:
+            return pathlib.Path(license_.file)
         return None
 
     @property
@@ -913,7 +874,7 @@ class Project():
         meson_dist_path = pathlib.Path(self._build_dir, 'meson-dist', f'{meson_dist_name}.tar.gz')
         sdist = pathlib.Path(directory, f'{dist_name}.tar.gz')
 
-        with tarfile.open(meson_dist_path, 'r:gz') as meson_dist, mesonpy._util.create_targz(sdist) as (tar, mtime):
+        with tarfile.open(meson_dist_path, 'r:gz') as meson_dist, mesonpy._util.create_targz(sdist) as tar:
             for member in meson_dist.getmembers():
                 # calculate the file path in the source directory
                 assert member.name, member.name
@@ -940,6 +901,7 @@ class Project():
 
                 info = tarfile.TarInfo(member.name)
                 file_stat = os.stat(path)
+                info.mtime = member.mtime
                 info.size = file_stat.st_size
                 info.mode = int(oct(file_stat.st_mode)[-3:], 8)
 
@@ -955,8 +917,7 @@ class Project():
 
             # add PKG-INFO to dist file to make it a sdist
             pkginfo_info = tarfile.TarInfo(f'{dist_name}/PKG-INFO')
-            if mtime:
-                pkginfo_info.mtime = mtime
+            pkginfo_info.mtime = time.time()  # type: ignore[assignment]
             pkginfo_info.size = len(self.metadata)
             tar.addfile(pkginfo_info, fileobj=io.BytesIO(self.metadata))
 
@@ -975,18 +936,19 @@ class Project():
 
 
 @contextlib.contextmanager
-def _project(config_settings: Optional[Dict[Any, Any]]) -> Iterator[Project]:
+def _project(config_settings: Optional[Dict[Any, Any]] = None) -> Iterator[Project]:
     """Create the project given the given config settings."""
 
     settings = _validate_config_settings(config_settings or {})
-    meson_args = {name: settings.get(f'{name}-args', []) for name in _MESON_ARGS_KEYS}
+    meson_args = typing.cast(MesonArgs, {name: settings.get(f'{name}-args', []) for name in _MESON_ARGS_KEYS})
+    source_dir = os.path.curdir
+    build_dir = settings.get('builddir')
+    editable_verbose = bool(settings.get('editable-verbose'))
 
-    with Project.with_temp_working_dir(
-            build_dir=settings.get('builddir'),
-            meson_args=typing.cast(MesonArgs, meson_args),
-            editable_verbose=bool(settings.get('editable-verbose'))
-    ) as project:
-        yield project
+    with contextlib.ExitStack() as ctx:
+        if build_dir is None:
+            build_dir = ctx.enter_context(tempfile.TemporaryDirectory(prefix='.mesonpy-', dir=source_dir))
+        yield Project(source_dir, build_dir, meson_args, editable_verbose)
 
 
 def _parse_version_string(string: str) -> Tuple[int, ...]:
